import builtins
import logging
import statistics
from collections import OrderedDict

from hydra_genetics.utils.io.chr import ChrTranslater
from hydra_genetics.utils.models.hotspot import MultiBpVariantData
from hydra_genetics.utils.models.hotspot import ReportClass
from hydra_genetics.utils.io.hotspot import Reader as HotspotReader
from hydra_genetics.utils.io import utils

from pysam import VariantFile

import yaml

log = logging.getLogger()


def generate_hotspot_report(sample,
                            output,
                            levels,
                            hotspot_file,
                            vcf_file,
                            gvcf_file,
                            chr_mapping,
                            vcf_file_wo_pick=None,
                            column_yaml_file=None):
    reports = OrderedDict(((ReportClass.hotspot, {}),
                          (ReportClass.region_all, {}),
                          (ReportClass.region, {}),
                          (ReportClass.indel, {'indel': []})))
    chr_translater = ChrTranslater(chr_mapping)

    chromosomes_to_look_at = set()
    if not hotspot_file == "-":
        try:
            hotspot_reader = HotspotReader(hotspot_file)
            for hotspot in iter(hotspot_reader):
                chromosomes_to_look_at.add(chr_translater.get_chr_value(hotspot.CHROMOSOME))
                if hotspot.REPORT == ReportClass.indel:
                    reports[hotspot.REPORT]['indel'].append(hotspot)
                else:
                    key = f"{hotspot.CHROMOSOME}-{hotspot.START // 1000000}"
                    reports[hotspot.REPORT][key] = reports[hotspot.REPORT].get(key, []) + [hotspot]
        except ValueError as e:
            logging.error(e)
            exit(1)

    variants = VariantFile(vcf_file)
    other = []

    transcript_dict = {}
    log.info("Processing vep transcripts")
    for record in variants.header.records:
        if record.type == "INFO":
            if record['ID'] == "CSQ":
                vep_fields = {v: c for c, v in enumerate(record['Description'].split("Format: ")[1].split('">')[0].split("|"))}

    for variant in variants:
        if variant is None:
            raise Exception("Empty allele found: " + str(variant))
        if not len(variant.alts) == 1:
            raise Exception("Multiple allele found: " + str(variant.alts))
        chromosomes_to_look_at.add(variant.chrom)
        variant_key = f"{variant.chrom}_{variant.start}_{variant.stop}_{variant.ref}_{','.join(variant.alts)}"
        try:
            transcript = variant.info['CSQ'][0].split("|")[vep_fields['Feature']]
        except KeyError:
            continue
        transcript_dict[variant_key] = transcript

    if vcf_file_wo_pick is not None:
        variants = VariantFile(vcf_file_wo_pick)
    else:
        variants = VariantFile(vcf_file)
    log.info("Processing variants")
    sub_report_keys = list(reports.keys())[0:-1]
    indel_report_key = list(reports.keys())[-1]
    for variant in variants:
        # ToDo make sure that empty variants are handled better!!!
        if variant is None:
            raise Exception("Empty allele found: " + str(variant))
        if not len(variant.alts) == 1:
            raise Exception("Multiple allele found: " + str(variant.alts))
        variant_key = f"{variant.chrom}_{variant.start}_{variant.stop}_{variant.ref}_{','.join(variant.alts)}"
        variant_key_hotspot = f"{chr_translater.get_nc_value(variant.chrom)}-{variant.start // 1000000}"
        if variant_key in transcript_dict:
            added = False
            for report in sub_report_keys:
                for hotspot in reports[report].get(variant_key_hotspot, []):
                    if hotspot.add_variant(variant, chr_translater):
                        hotspot_transcript = hotspot.ACCESSION_NUMBER
                        if not hotspot_transcript == "-":
                            transcript_dict[variant_key] = hotspot_transcript
                        log.debug("Adding variant {}:{}-{} {} {} to hotspot: {}".format(variant.chrom,
                                                                                        variant.start,
                                                                                        variant.stop,
                                                                                        variant.ref,
                                                                                        ",".join(variant.alts),
                                                                                        hotspot))
                        added = True
                        break
                if added:
                    break
            if not added:
                for hotspot in reports[indel_report_key]['indel']:
                    if hotspot.add_variant(variant, chr_translater):
                        hotspot_transcript = hotspot.ACCESSION_NUMBER
                        if not hotspot_transcript == "-":
                            transcript_dict[variant_key] = hotspot_transcript
                        log.debug("Adding variant {}:{}-{} {} {} to hotspot: {}".format(variant.chrom,
                                                                                        variant.start,
                                                                                        variant.stop,
                                                                                        variant.ref,
                                                                                        ",".join(variant.alts),
                                                                                        hotspot))
                        added = True
                        break
            if not added:
                chromosomes_to_look_at.add(variant.chrom)
                other.append(variant)
    log.info("Open genomic vcf")
    g_variants = VariantFile(gvcf_file)

    columns = {'columns': []}
    if column_yaml_file is not None:
        log.info("Process yaml for: {}".format(column_yaml_file))
        with open(column_yaml_file) as file:
            columns = yaml.load(file, Loader=yaml.FullLoader)

    output_order = []
    hotspot_columns = {'sample': {},
                       'chr': {},
                       'start': {},
                       'stop': {},
                       'ref': {},
                       'alt': {},
                       'report': {},
                       'gvcf_depth': {},
                       'ref_depth': {},
                       'alt_depth': {}}
    report_header = []

    biggest_order = -1

    gcvf_depth_field = 'DP'
    if isinstance(columns['columns'], dict):
        gcvf_depth_field = columns['columns'].get('gvcf_depth', {}).get('field', 'DP')
    for entry in columns['columns']:
        if entry in hotspot_columns:
            if columns['columns'][entry].get('visible', 1) == 0 or 'order' in columns['columns'][entry]:
                del hotspot_columns[entry]
            elif columns['columns'][entry].get('format', None):
                hotspot_columns[entry]['format'] = columns['columns'][entry]['format']
        if 'order' in columns['columns'][entry]:
            output_order.append((columns['columns'][entry]['order'], entry))
            if 'from' in columns['columns'][entry] and 'merge' == columns['columns'][entry]['from']:
                report_header.append((columns['columns'][entry]['order'],
                                      entry + "_" +
                                      columns['columns'][entry]['divider'].join(
                                        extract_item_merge_header(columns['columns'][entry]['elements']))))
            else:
                report_header.append((columns['columns'][entry]['order'], entry))
            biggest_order = max(biggest_order, columns['columns'][entry]['order'])
    for key in hotspot_columns:
        if key in columns['columns']:
            del columns['columns'][key]

    depth_data = dict(map(lambda x: (x, {}), chromosomes_to_look_at))

    for chr in depth_data:
        for variant in g_variants.fetch(chr):
            depth_data[chr][variant.start] = variant.samples[sample][gcvf_depth_field]

    def handle_select(data):
        def convert_list_slice(info):
            if len(info) < 3:
                raise SyntaxError(f"Invalid syntax: {info}")
            if info.startswith("[") and info.endswith("]"):
                info = info[1:-1].split(":")
                if len(info) == 1:
                    return lambda data: data[int(info[0])]
                elif len(info) == 2:
                    value0 = int(info[0]) if len(info[0]) > 0 else None
                    value1 = int(info[1]) if len(info[1]) > 0 else None
                    return lambda data: data[value0:value1]
                elif len(info) == 3:
                    value0 = int(info[0]) if len(info[0]) > 0 else None
                    value1 = int(info[1]) if len(info[1]) > 0 else None
                    value2 = int(info[2]) if len(info[2]) > 0 else None
                    return lambda data: data[value0:value1:value2]
                else:
                    raise SyntaxError(f"Invalid syntax: {info}, invalid number of elements inside '[]'")
            else:
                raise SyntaxError(f"Invalid syntax: {info}, missing bracket '[]'")

        def condition(items, empty):
            items = convert_list_slice(items)

            def func(data):
                values = items(data)
                if isinstance(values, str):
                    return (values == "-") == empty
                else:
                    return (len([v for _, v in values.items() if v != "-"]) == 0) == empty
            return func

        data['items'] = convert_list_slice(data['items'])
        if "else" in data:
            data['else'] = convert_list_slice(data['else'])
        if "condition" in data:
            data['condition'] = condition(data['condition']['items'], data['condition'].get("empty", True))
        return data

    for entry in columns['columns']:
        if "from" in columns['columns'][entry] and "select" in columns['columns'][entry]["from"]:
            columns['columns'][entry] = handle_select(columns['columns'][entry])

    for entry in hotspot_columns:
        biggest_order += 1
        output_order.append((biggest_order, entry))
        report_header.append((biggest_order, entry))
    for entry in columns['columns']:
        if 'order' not in columns['columns'][entry]:
            if columns['columns'][entry].get('visible', 1) == 0:
                continue
            biggest_order += 1
            output_order.append((biggest_order, entry))
            if 'from' in columns['columns'][entry] and 'merge' == columns['columns'][entry]['from']:
                report_header.append((biggest_order,
                                      entry + "_" +
                                      columns['columns'][entry]['divider'].join(
                                        extract_item_merge_header(columns['columns'][entry]['elements']))))
            else:
                report_header.append((biggest_order, entry))
    output_order = sorted(output_order, key=lambda tup: tup[0])
    report_header = sorted(report_header, key=lambda tup: tup[0])

    log.info("Process vcf header: {}".format(vcf_file))
    for record in variants.header.records:
        if record.type == "INFO":
            if record['ID'] == "CSQ":
                log.info(" -- found vep information: {}".format(vcf_file))
                log.debug(" -- -- {}".format(record['Description'].split("Format: ")[1].split('">')[0].split("|")))
                vep_fields = {v: c for c, v in enumerate(record['Description'].split("Format: ")[1].split('">')[0].split("|"))}
                annotation_extractor = utils.get_annotation_data_vep(vep_fields, transcript_dict)

    def get_depth(data, chr, start, stop):
        depth = [data[chr].get(pos, 0) for pos in range(start, stop)]
        if len(depth) > 1:
            return statistics.mean(depth)
        elif len(depth) == 1:
            return depth[0]
        else:
            return 0

    log.info("open output file: {}".format(output))
    with open(output, "w") as writer:
        writer.write("\t".join([name[1] for name in report_header]))
        log.info("Printing header: {}".format(output))
        log.info("Printing hotspot information: {}".format(output))
        counter = 0
        for report in reports:
            for index_data in reports[report]:
                for hotspot in reports[report][index_data]:
                    for index, variant in enumerate(hotspot.VARIANTS):
                        # even though no variants were found print hotspot and region all entries
                        if not variant['variants'] and not variant['extended']:
                            depth = get_depth(depth_data,
                                              chr_translater.get_chr_value(hotspot.CHROMOSOME),
                                              hotspot.EXTENDED_START + index-1,
                                              hotspot.EXTENDED_START + index)
                            if hotspot.ALWAYS_PRINT:
                                data = {'sample': sample,
                                        'chr': hotspot.CHROMOSOME,
                                        'start': hotspot.EXTENDED_START + index,
                                        'stop': hotspot.EXTENDED_START + index,
                                        'ref': '-',
                                        'alt': '-',
                                        'report':  utils.format_report_type(hotspot),
                                        'gvcf_depth': depth,
                                        'ref_depth': '-',
                                        'alt_depth': '-'}
                                format_hotspot(data, columns, hotspot_columns)
                                add_columns(data, None, hotspot, columns, annotation_extractor, depth, levels)
                                writer.write("\n" + "\t".join([str(data[c[1]]) for c in output_order]))
                                counter += 1
                        else:
                            # print found variants that overlap with hotspot positions
                            for var in variant['variants']:
                                depth = get_depth(depth_data,
                                                  var.chrom,
                                                  var.start,
                                                  var.stop)
                                data = {'sample': sample,
                                        'chr': chr_translater.get_nc_value(var.chrom),
                                        'start': var.start + 1,
                                        'stop': var.stop,
                                        'ref': var.ref,
                                        'alt': ",".join(var.alts),
                                        'report': utils.get_report_type(var, hotspot),
                                        'gvcf_depth': depth,
                                        'ref_depth': var.samples[sample]['AD'][0],
                                        'alt_depth': ",".join(map(str, var.samples[sample]['AD'][1:]))}
                                format_hotspot(data, columns, hotspot_columns)
                                add_columns(data, var, hotspot, columns, annotation_extractor, depth, levels)
                                writer.write("\n" + "\t".join([str(data[c[1]]) for c in output_order]))
                                counter += 1
        log.info("-- hotspot entries: {}".format(counter))
        log.info("Printing variants that aren't hotspot: {}".format(output))
        counter = 0
        for var in other:
            # print variants that doesn't overlap with a hotspot
            depth = get_depth(depth_data,
                              var.chrom,
                              var.start,
                              var.stop)
            data = {'sample': sample,
                    'chr': chr_translater.get_nc_value(var.chrom),
                    'start': var.start + 1,
                    'stop': var.stop,
                    'ref': var.ref,
                    'alt': ",".join(var.alts),
                    'report': "4-other",
                    'gvcf_depth': depth,
                    'ref_depth': var.samples[sample]['AD'][0],
                    'alt_depth': ",".join(map(str, var.samples[sample]['AD'][1:]))}
            format_hotspot(data, columns, hotspot_columns)
            add_columns(data, var, None, columns, annotation_extractor, depth, levels)
            writer.write("\n" + "\t".join([str(data[c[1]]) for c in output_order]))
            counter += 1
            log.info("-- non-hotspot entries: {}".format(counter))


def format_value(value, format):
    if format[0] == "replace":
        return value.replace(format[1], format[2])
    elif format[0] == "string":
        if len(format) == 3:
            value = getattr(builtins, format[2])(value)
        else:
            value = float(value)
        return format[1].format(value)
    else:
        raise Exception("Unknown format value: " + format)


def format_hotspot(data, columns, hotspot):
    for key in data:
        format = columns.get(key, {}).get("format", None)
        format = hotspot.get(key, {}).get("format", format)
        if format:
            try:
                data[key] = format_value(data[key], format)
            except ValueError:
                log.warning("Unable to format value {}, field {}, format {}".format(data[key], key, format))
            except TypeError:
                log.warning("Unable to format value {}, field {}, format {}".format(data[key], key, format))


def extract_item_merge_header(columns):
    header_list = []
    for value in columns:
        if 'from' in columns[value] and columns[value]['from'] == 'merge':
            header_list.append(value + "_" +
                               columns[value]['divider'].join(extract_item_merge_header(columns[value]['elements'])))
        else:
            header_list.append(value)
    return header_list


def add_columns(data, var, hotspot, columns, annotation_extractor, depth, levels):
    """
        print extra columns defined by a yaml file. Data from variant, hotspot or annotations
    """
    def add_data(data, column, c, depth, levels):
        if column[c]['from'] == "merge":
            # Used to merge two columns and combine using divider
            divider = column[c]['divider']
            temp_data = OrderedDict()
            for key in column[c]['elements']:
                add_data(temp_data, column[c]['elements'], key, depth, levels)
            data[c] = divider.join([v for k, v in temp_data.items()])
        elif column[c]['from'] == "select":
            # From multiple column select the first one with data
            temp_data = OrderedDict()
            for key in column[c]['elements']:
                add_data(temp_data, column[c]['elements'], key, depth, levels)
            temp_data = [v for _, v in temp_data.items()]
            select = column[c]['items'](temp_data)
            if not isinstance(select, str):
                select = column[c].get("divider", ":").join(select)
            data[c] = "-"
            condition = True
            if 'condition' in column[c]:
                condition = column[c]['condition'](temp_data)
            if select is not None and select != '-' and condition:
                data[c] = select
            elif 'else' in column[c]:
                select = column[c]['else'](temp_data)
                if not isinstance(select, str):
                    select = [v for v in select if v != "-"]
                    if len(select) == 0:
                        select = "-"
                    else:
                        select = column[c].get("divider", ":").join(select)
                data[c] = select
        elif column[c]['from'] == "vep":
            try:
                data[c] = annotation_extractor(var,  column[c]['field'])
                if data[c] is None or data[c] == "":
                    data[c] = '-'
            except AttributeError:
                data[c] = '-'
            if "extract_regex" in column[c]:
                data[c] = utils.regex_extract(data[c], column[c]['extract_regex'])
        elif column[c]['from'] == "hotspot":
            if hotspot is None:
                data[c] = "-"
            else:
                data[c] = getattr(hotspot, column[c]['field'])
        elif column[c]['from'] == "function":
            function = getattr(utils, column[c]['name'])
            variables = ()
            if 'variables' in column[c]:
                variables = []
                for v in column[c]['variables']:
                    variables.append(locals().get(v, v))
                try:
                    value = function(*variables)
                except AttributeError:
                    value = "-"
            else:
                value = function()
            if 'column' in column[c]:
                data[c] = value[column[c]['column']]
            else:
                if isinstance(value, tuple):
                    value = ",".join(value)
                data[c] = value
            if data[c] is None:
                data[c] = "-"
        elif column[c]['from'] == 'variable':
            data[c] = locals()[column[c]['field']]
        else:
            raise Exception("Undhandledd cased: " + column[c]['field'])
        if "format" in column[c]:
            try:
                data[c] = format_value(data[c], column[c]["format"])
            except ValueError:
                log.debug("Unable to format value {}, field {}, format {}".format(data[c], c, column[c]["format"]))
            except TypeError:
                log.debug("Unable to format value {}, field {}, format {}".format(data[c], c, column[c]["format"]))

    for c in columns['columns']:
        if 'from' in columns['columns'][c]:
            add_data(data, columns['columns'], c, depth, levels)



#!/usr/bin/env python
"""Creates a hydra-core pipeline matching the current
organization's specification based on a template.
"""

from datetime import date
import git
import glob
import gzip
import logging
import jinja2
import json
import pathlib
import os
import re
import pandas as pd
import pysam
import shutil
import sys
import hydra_genetics

log = logging.getLogger(__name__)


class PipelineCreate(object):
    """Creates a hydra-genetics pipeline.
    Args:
        name (str): Name for the pipeline.
        description (str): Description for the pipeline.
        author (str): Authors name
        email (str): Authors email
        version (str): Version flag. Semantic versioning only. Defaults to `1.0dev`.
        snakemake_version (str): min snakemake version
        git_user: (str) username on github
        no_git (bool): Prevents the creation of a local Git repository for the pipeline. Defaults to False.
        force (bool): Overwrites a given workflow directory with the same name. Defaults to False.
            May the force be with you.
        outdir (str): Path to the local output directory.
    """
    def __init__(self, name, description, author, email, version="0.0.1", min_snakemake_version="6.10.0",
                 git_user=None, no_git=False, force=False, outdir=None):
        self.short_name = name.lower().replace(r"/\s+/", "-").replace("hydra-genetics/", "").replace("/", "-")
        self.name = f"{self.short_name}"
        self.description = description
        self.author = author
        self.email = email
        self.version = version
        self.min_snakemake_version = min_snakemake_version
        self.git_user = git_user
        self.no_git = no_git
        self.force = force
        self.year = date.today().year
        self.outdir = outdir
        if not self.outdir:
            self.outdir = os.path.join(os.getcwd(), self.name)

    def init_pipeline(self):
        """Creates the hydra_genetics pipeline."""
        log.info(f"Creating pipeline: {self.short_name}")
        if os.path.exists(self.outdir):
            if self.force:
                log.warning(f"Output directory '{self.outdir}' exists - continuing as --force specified")
            else:
                log.error(f"Output directory '{self.outdir}' exists!")
                log.info("Use -f / --force to overwrite existing files")
                sys.exit(1)
        else:
            os.makedirs(self.outdir)

        env = jinja2.Environment(
           loader=jinja2.PackageLoader("hydra_genetics", "pipeline-template"), keep_trailing_newline=True
        )

        template_dir = os.path.join(os.path.dirname(__file__), "../pipeline-template")
        object_attrs = vars(self)
        template_files = list(pathlib.Path(template_dir).glob("**/*"))
        ignore_strs = [".pyc", "__pycache__", ".pyo", ".pyd", ".DS_Store", ".egg", ".snakemake"]
        log.info("Adding folder and files")
        for template_fn_path_obj in template_files:
            template_fn_path = str(template_fn_path_obj)
            if any([s in template_fn_path for s in ignore_strs]):
                log.debug(f"Ignoring '{template_fn_path}' in jinja2 template creation")
                continue

            template_fn = os.path.relpath(template_fn_path, template_dir)
            output_path = os.path.join(self.outdir, template_fn)

            if os.path.isdir(template_fn_path_obj):
                log.debug(f"Creating directory: '{output_path}'")
                if os.path.exists(output_path):
                    if self.force:
                        log.warning(f"Output directory '{output_path}' exists - continuing as --force specified")
                    else:
                        log.error(f"Output directory '{output_path}' exists!")
                        log.info("Use -f / --force to overwrite existing files")
                        sys.exit(1)
                else:
                    os.makedirs(output_path)
                continue

            log.debug(f"Rendering template file: '{template_fn}'")

            if template_fn.endswith(".png"):
                shutil.copy(os.path.join(template_dir, template_fn), output_path)
            else:
                j_template = env.get_template(template_fn)
                rendered_output = j_template.render(object_attrs)

                # Write to the pipeline output file
                with open(output_path, "w") as fh:
                    log.debug(f"Writing to output file: '{output_path}'")
                    fh.write(rendered_output)

        if not self.no_git:
            self.git_init_pipeline()

    def git_init_pipeline(self):
        """Initialises the new pipeline as a Git repository and submits first commit."""
        log.info("Initialising pipeline git repository")
        repo = git.Repo.init(self.outdir, initial_branch='main')
        repo.git.add(A=True)
        repo.index.commit(f"initial template build from hydra-genetics/tools, version {hydra_genetics.__version__}")
        # Add TEMPLATE branch to git repository
        repo.git.branch("develop")
        log.info(
            "Done. Remember to add a remote and push to GitHub:\n"
            f"[white on grey23] cd {self.outdir} \n"
            " git remote add origin git@github.com:USERNAME/REPO_NAME.git \n"
            " git push --all origin                                       "
        )


class RuleCreate(object):
    """Creates a hydra-genetics rule.
    Args:
        command (str): Name of the rule, command that will be run.
        tool (str): tool that will be used to run the command, ex samtools (optional)
        module (str): Name of module where rule will be added.
        author (str): Authors name
        email (str): Authors email
        outdir (str): Path to the local output directory.

        if a tool and command are both provided the rule will be named tool_command
    """
    def __init__(self, command, module, author, email, tool=None, outdir=None):
        self.command = command
        self.tool = tool
        self.module_name = module
        self.author = author
        self.email = email
        self.year = date.today().year

        if not outdir:
            self.outdir = os.getcwd()
        else:
            self.outdir = os.path.abspath(outdir)

    def init_rule(self):
        """Creates the hydra_genetics rule."""
        outdir = os.path.normpath(os.path.join(self.outdir, self.module_name))
        if not os.path.exists(outdir):
            outdir_temp = outdir
            outdir = os.path.join(os.path.dirname(self.outdir), self.module_name)
            if not os.path.exists(outdir):
                log.error(f"Can not find module directory, tried with:\n'{outdir_temp}'\n'{outdir}'")
                exit(1)
        outdir_workflow = os.path.join(outdir, "workflow")
        # Rule path
        output_rules = os.path.join(outdir_workflow, "rules")

        # Schema path
        output_schemas = os.path.join(outdir_workflow, "schemas")
        output_schemas_config = os.path.join(output_schemas, "config.schema.yaml")
        output_schemas_resources = os.path.join(output_schemas, "resources.schema.yaml")
        output_schemas_rule = os.path.join(output_schemas, "rules.schema.yaml")

        # Docs software path
        output_docs_software = os.path.join(outdir, "docs", "softwares.md")

        if not os.path.exists(output_rules):
            log.error(f"Can not find output directory '{output_rules}'")
            sys.exit(2)
        self.append_rule = False
        self.name = self.command
        if self.tool is None:
            self.filename_rulename = f"{self.name}__{self.name}"
            log.info(f"Creating rule: '{output_rules}/{self.name}.smk'")
            output_rule = os.path.join(output_rules, f"{self.name}.smk")
        else:
            self.name = f"{self.tool}_{self.name}"
            self.filename_rulename = f"{self.tool}__{self.name}"
            if os.path.exists(os.path.join(output_rules, f"{self.tool}.smk")):
                log.info(f"Adding entry {self.name} to smk '{outdir}/{self.tool}.smk'")
                output_rule = os.path.join(output_rules, f"{self.tool}.smk")
                self.append_rule = True
            else:
                log.info(f"Creating smk file: '{output_rules}/{self.tool}.smk' with rule {self.name}")
                output_rule = os.path.join(output_rules, f"{self.tool}.smk")
        if os.path.exists(output_rule) and self.tool is None:
            log.error(f"Rule already exists '{output_rule}'")
            sys.exit(4)

        if not os.path.exists(output_schemas):
            log.info(f"Creating schema folder {output_schemas}")
            os.mkdir(output_schemas)

        if not os.path.exists(output_schemas_config):
            log.info(f"Creating config schema file {output_schemas_config}")
            self.append_schema_config = False
        else:
            log.info(f"Appending to config schema file {output_schemas_config}")
            self.append_schema_config = True

        if not os.path.exists(output_schemas_resources):
            log.info(f"Creating resources schema file {output_schemas_resources}")
            self.append_schema_resources = False
        else:
            log.info(f"Appending to resources schema file {output_schemas_resources}")
            self.append_schema_resources = True

        if not os.path.exists(output_schemas_rule):
            log.info(f"Creating rule schema file {output_schemas_rule}")
            self.append_schema_rule = False
        else:
            log.info(f"Appending to rule schema file {output_schemas_rule}")
            self.append_schema_rule = True

        if not os.path.exists(output_docs_software):
            log.info(f"Creating docs file {output_docs_software}")
            self.append_docs_software = False
            parent_dir = os.path.dirname(output_docs_software)
            if not os.path.exists(parent_dir):
                os.mkdir(parent_dir)
        else:
            log.info(f"Appending to docs file {output_docs_software}")
            self.append_docs_software = True

        env = jinja2.Environment(
            loader=jinja2.PackageLoader("hydra_genetics", "rule-template"), keep_trailing_newline=True
        )
        template_dir = os.path.join(os.path.dirname(__file__), "../rule-template")
        rename_files = {
            "skeleton_rule.smk": output_rule,
        }
        object_attrs = vars(self)
        template_files = list(pathlib.Path(template_dir).glob("**/*"))
        ignore_strs = [".pyc", "__pycache__", ".pyo", ".pyd", ".DS_Store", ".egg", ".snakemake"]
        if self.append_rule:
            with open(output_rule, 'r') as lines:
                for line in lines:
                    if self.name in line:
                        log.error(f"rule {self.name} already found in file {output_rule}")
                        exit(6)
        log.info("Adding folder and files")
        for template_fn_path_obj in template_files:
            template_fn_path = str(template_fn_path_obj)
            if any([s in template_fn_path for s in ignore_strs]):
                log.debug(f"Ignoring '{template_fn_path}' in jinja2 template creation")
                continue

            template_fn = os.path.relpath(template_fn_path, template_dir)
            if template_fn.endswith(".smk"):
                output_path = output_rules
                self.append = self.append_rule
            elif template_fn.endswith("config.schema.yaml"):
                output_path = output_schemas_config
                self.append = self.append_schema_config
            elif template_fn.endswith("rules.schema.yaml"):
                output_path = output_schemas_rule
                self.append = self.append_schema_rule
            elif template_fn.endswith("resources.schema.yaml"):
                output_path = output_schemas_resources
                self.append = self.append_schema_resources
            elif template_fn.endswith(".md"):
                output_path = output_docs_software
                self.append = self.append_docs_software

            if template_fn in rename_files:
                output_path = os.path.join(outdir, rename_files[template_fn])

            log.info(f"Rendering template file: '{template_fn}' at {output_path}")
            j_template = env.get_template(template_fn)
            rendered_output = j_template.render(object_attrs, trim_blocks=False)

            if os.path.exists(output_path):
                with open(output_path, "r+") as fh:
                    lines = fh.readlines()
                    line_number = 0
                    for line in lines:
                        if line.startswith('required:'):
                            break
                        else:
                            line_number += 1
                    fh.seek(0)
                    lines.insert(line_number, rendered_output)
                    log.info(f"Appending to output file: '{output_path}'")
                    fh.writelines(lines)
            else:
                # Write to the content to a new file
                with open(output_path, "w") as fh:
                    log.info(f"Writing to new output file: '{output_path}'")
                    fh.write(rendered_output)
        snakefile = os.path.join(outdir_workflow, "Snakefile")
        with open(snakefile, 'r+') as fh:
            lines = fh.readlines()
            line_number = 0
            for line in lines:
                line_number += 1
                if line.startswith('include: "rules/common.smk"'):
                    break
            else:
                log.error(f"Could not find 'include: ../rules/common.smk' entry in Snakefile '{snakefile}'")
                sys.exit(1)
            fh.seek(0)
            if self.tool is not None and not self.append_rule:
                lines.insert(line_number, str(f"include: \"rules/{self.tool}.smk\"\n"))
            elif self.tool is None:
                lines.insert(line_number, str(f"include: \"rules/{self.command}.smk\"\n"))

            fh.writelines(lines)


class CreateInputFiles(object):
    """Creates a hydra-genetics input files samples.tsv and units.tsv.
    Args:
        directories (str): Path to dir that should be search.
        outdir (str): Path to the local output directory.
    """
    def __init__(self,
                 directory,
                 outdir=None,
                 post_file_modifier=None,
                 platform="Illumina",
                 sample_type="T",
                 sample_regex=r"^([A-Za-z0-9-]+)_.*\.fastq.gz",
                 read_number_regex="_(R[12]{1})[_.]{1}",
                 adapters="AGATCGGAAGAGCACACGTCTGAACTCCAGTCA,AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGT",
                 data_json=None,
                 data_columns=None,
                 tc=1.0,
                 force=False,
                 default_barcode=None,
                 validate_run_information=False,
                 ask_for_input=False,
                 occurrences_warning_th=0.9,
                 number_of_reads=200,
                 every_n_reads=1000):
        self.directory = directory
        self.outdir = outdir
        self.post_file_modifier = post_file_modifier
        self.platform = platform
        self.sample_type = sample_type
        self.sample_regex = sample_regex
        self.read_number_regex = read_number_regex
        self.adapters = adapters
        self.data_json = data_json
        self.data_columns = data_columns
        self.tc = tc
        self.force = force
        self.default_barcode = default_barcode
        self.validate_run_information = validate_run_information
        self.ask_for_input = ask_for_input
        self.occurrences_warning_th = occurrences_warning_th
        self.number_of_reads = number_of_reads
        self.every_n_reads = every_n_reads

        if not self.outdir:
            self.outdir = os.getcwd()

    def init(self):
        """Creates the hydra_genetics pipeline input files."""
        log.info(f"Searching for fastq-files:")
        file_dict = {}
        data_json = {}
        data_columns = {}
        if self.data_json and self.data_columns:
            data_json = json.load(open(self.data_json))
            data_columns = json.load(open(self.data_columns))
            if 'units' in data_columns:
                data_columns['units'] = dict(map(lambda v: (v.split(":")[0].split(".")[-1],
                                                            (v.split(":")[1], v.split(":")[0].split("."))),
                                                 data_columns['units']))
            if 'samples' in data_columns:
                data_columns['samples'] = dict(map(lambda v: (v.split(":")[0].split(".")[-1],
                                                              (v.split(":")[1], v.split(":")[0].split('.'))),
                                                   data_columns['samples']))
        elif self.data_json or self.data_columns:
            log.error("Both --data-json and --data-columns need to be specified at the same time, not only one of them.")
        for d in self.directory:
            dir_files_found = 0
            log.info(f"Dir: %s" % d)
            for f in glob.glob('%s/**/*.fastq.gz' % d, recursive=True):
                if "Undetermined" in f:
                    continue
                temp_filename = os.path.basename(f)
                try:
                    samplename = re.search(self.sample_regex, temp_filename).group(1)
                    read = re.search(self.read_number_regex, temp_filename).group(1)
                    first_part = re.split(self.read_number_regex, temp_filename)[0]
                    log.debug("sample name: {}, read part: {}".format(samplename, read))
                    if samplename not in file_dict:
                        file_dict[samplename] = {}
                    if first_part not in file_dict[samplename]:
                        file_dict[samplename][first_part] = {}
                    if read in file_dict[samplename][first_part]:
                        raise Exception(
                                    "Trying to add read (%s) for sample %s multiple times\n%s\n%s" %
                                    (read, samplename, f, file_dict[samplename][first_part][read]))
                    file_dict[samplename][first_part][read] = f
                    dir_files_found += 1
                    log.info("    found: %s" % f)
                except AttributeError:
                    log.debug("Couldn't extract sample name from: %s" % temp_filename)
            if dir_files_found == 0:
                log.warning("No fastq files found in '{}', "
                            "please make sure regex '{}' matches your file names!, and '{}' matches read number".
                            format(d, self.sample_regex, self.read_number_regex))
                exit(1)
            else:
                log.info("{} fastq files found".format(dir_files_found))
        result_dict = {}
        if self.validate_run_information:
            log.info("NOTE: fastq file will be parsed until end, could take some time for big files")
        log.info("Processing  found files, extracting run information and validation number of found reads:".format(str(f)))
        for sample in file_dict:
            for files in file_dict[sample]:
                if len(file_dict[sample][files]) % 2 != 0:
                    raise ValueError("Uneven number of files found:\n{}".format(str(file_dict[sample][files])))
                for read_number, f in file_dict[sample][files].items():
                    log.info("\t - {} for run information".format(str(f)))
                    machine_id, flowcell, lane_id, barcode = extract_run_information(f,
                                                                                     self.default_barcode,
                                                                                     self.number_of_reads,
                                                                                     self.every_n_reads,
                                                                                     self.occurrences_warning_th,
                                                                                     self.validate_run_information,
                                                                                     self.ask_for_input)
                    no_index_counter = 0
                    if sample in result_dict:
                        if flowcell in result_dict[sample]:
                            if barcode in result_dict[sample][flowcell]:
                                if lane_id in result_dict[sample][flowcell][barcode]:
                                    if 'reads' in result_dict[sample][flowcell][barcode][lane_id]:
                                        if read_number in result_dict[sample][flowcell][barcode][lane_id]['reads']:
                                            dup_file = result_dict[sample][flowcell][barcode][lane_id]['reads'][read_number]
                                            raise ValueError("sample, flowcell, lane and read number combination "
                                                             "found multiple times: sample {}"
                                                             " flowcell {} barcode {} lane {} read {}:\n - {}".
                                                             format(sample,
                                                                    flowcell,
                                                                    barcode,
                                                                    lane_id,
                                                                    read_number,
                                                                    "\n - ".join([f, dup_file])))
                                        else:
                                            result_dict[sample][flowcell][barcode][lane_id]['reads'][read_number] = f
                                    else:
                                        result_dict[sample][flowcell][barcode][lane_id]['reads'] = {read_number: f}
                                        result_dict[sample][flowcell][barcode][lane_id]['machine'] = machine_id
                                        result_dict[sample][flowcell][barcode][lane_id]['barcode'] = barcode
                                else:
                                    result_dict[sample][flowcell][barcode][lane_id] = {'reads': {read_number: f},
                                                                                       'machine': machine_id,
                                                                                       'barcode': barcode}
                            else:
                                result_dict[sample][flowcell][barcode] = {lane_id: {'reads': {read_number: f},
                                                                                    'machine': machine_id,
                                                                                    'barcode': barcode}}
                        else:
                            result_dict[sample][flowcell] = {barcode: {lane_id: {'reads': {read_number: f},
                                                                                 'machine': machine_id,
                                                                                 'barcode': barcode}}}
                    else:
                        result_dict[sample] = {flowcell: {barcode: {lane_id: {'reads': {read_number: f},
                                                                              'machine': machine_id,
                                                                              'barcode': barcode}}}}

        samples_file_name = "samples.tsv"
        if self.post_file_modifier is not None:
            samples_file_name = "samples_{}.tsv".format(self.post_file_modifier)
        if os.path.isfile(samples_file_name):
            if not self.force:
                log.warn("File exists {} and force wasn't used".format(samples_file_name))
                exit(1)
            else:
                log.warn("File exists {} overwriting!!!".format(samples_file_name))
        with open(samples_file_name, "w") as output:
            header = ["sample"]
            if data_columns and "samples" in data_columns:
                header += data_columns["samples"].keys()
            elif self.tc is not None:
                header.append('tumor_content')
            output.write("\t".join(header))
            for sample, data in sorted(file_dict.items()):
                row_data = [sample]
                if data_columns and "samples" in data_columns:
                    for _, value in data_columns['samples'].items():
                        row_data.append(extract_value(value[1][2:], value[0], data_json['samples'][sample]))
                elif self.tc is not None:
                    row_data.append(self.tc)
                output.write("\n{}".format("\t".join(map(lambda x: str(x), row_data))))
        units_file_name = "units.tsv"
        if self.post_file_modifier is not None:
            units_file_name = "units_{}.tsv".format(self.post_file_modifier)
        if os.path.isfile(units_file_name):
            if not self.force:
                log.warn("File exists {} and force wasn't used".format(units_file_name))
                exit(1)
            else:
                log.warn("File exists {} overwriting!!!".format(units_file_name))

        with open(units_file_name, "w") as output:
            extra_header_columns = []
            if 'units' in data_columns:
                extra_header_columns = list(data_columns['units'].keys())
                if 'type' in extra_header_columns:
                    extra_header_columns.remove("type")
                if 'platform' in extra_header_columns:
                    extra_header_columns.remove("platform")
                if 'adapter' in extra_header_columns:
                    extra_header_columns.remove("adapter")
            output.write("\t".join(["sample", "type", "platform", "barcode", "machine",
                                    "flowcell", "lane", "fastq1", "fastq2", "adapter"] + extra_header_columns))
            for sample in sorted(result_dict):
                for flowcell in sorted(result_dict[sample]):
                    for barcode in sorted(result_dict[sample][flowcell]):
                        for lane, data in sorted(result_dict[sample][flowcell][barcode].items()):
                            if len(data['reads'].keys()) != 2:
                                raise ValueError("Incorrect number of fastq-files: {}:\n - {}".format(
                                    len(data['reads'].keys()), "\n - ".join(
                                        "{}: {}".format(k, data['reads'][k]) for k in data['reads'])))
                            sample_type = self.sample_type
                            s_type = sample_type
                            s_platform = self.platform
                            s_adapters = self.adapters
                            extra_data = []
                            if data_json and 'units' in data_columns:
                                extra_columns = list(data_columns['units'].keys())
                                if 'type' in extra_columns:
                                    value = data_columns['units']['type']
                                    s_type = extract_value(value[1][2:], value[0], data_json['samples'][sample])
                                    extra_columns.remove('type')
                                if 'platform' in extra_columns:
                                    value = data_columns['units']['platform']
                                    s_platform = extract_value(value[1][2:], value[0], data_json['samples'][sample])
                                    extra_columns.remove('platform')
                                if 'adapter' in extra_columns:
                                    value = data_columns['units']['adapter']
                                    extra_columns.remove('adapter')
                                    s_adapters = extract_value(value[1][2:], value[0], data_json['samples'][sample])
                                for column in extra_columns:
                                    value = data_columns['units'][column]
                                    extra_data.append(extract_value(value[1][2:], value[0], data_json['samples'][sample]))
                            elif self.tc:
                                header.append('tumor_content')

                            output.write("\n"+"\t".join([sample,
                                                         s_type,
                                                         s_platform,
                                                         data['barcode'],
                                                         data['machine'],
                                                         flowcell,
                                                         "L" + lane.rjust(3, '0'),
                                                         str(data['reads']["1"]),
                                                         str(data['reads']["2"]),
                                                         s_adapters] + extra_data))


class CreateLongReadInputFiles(object):
    """Creates a hydra-genetics input files samples.tsv and units.tsv from
    unmapped BAM files containing long read sequence data.
    Args:
        directories (str): Path to dir that should be search.
        outdir (str): Path to the local output directory.
    """
    def __init__(self,
                 directory,
                 outdir=None,
                 post_file_modifier=None,
                 platform="ONT",
                 sample_type="N",
                 adapters=None,
                 data_json=None,
                 data_columns=None,
                 tc=None,
                 force=False,
                 default_barcode=None):
        self.directory = directory
        self.outdir = outdir
        self.post_file_modifier = post_file_modifier
        self.platform = platform
        self.sample_type = sample_type
        self.adapters = adapters
        self.data_json = data_json
        self.data_columns = data_columns
        self.tc = tc
        self.force = force
        self.default_barcode = default_barcode

        if not self.outdir:
            self.outdir = os.getcwd()

    def init(self):
        """Creates the hydra_genetics input files."""
        log.info("Searching for BAM-files:")
        data_json = {}
        data_columns = {}
        if self.data_json and self.data_columns:
            data_json = json.load(open(self.data_json))
            data_columns = json.load(open(self.data_columns))
            if 'units' in data_columns:
                data_columns['units'] = dict(map(lambda v: (v.split(":")[0].split(".")[-1],
                                                            (v.split(":")[1], v.split(":")[0].split("."))),
                                                 data_columns['units']))
            if 'samples' in data_columns:
                data_columns['samples'] = dict(map(lambda v: (v.split(":")[0].split(".")[-1],
                                                              (v.split(":")[1], v.split(":")[0].split('.'))),
                                                   data_columns['samples']))
        elif self.data_json or self.data_columns:
            log.error("Both --data-json and --data-columns need to be "
                      "specified at the same time, not only one of them.")

        sample_type = self.sample_type
        platform = self.platform
        adapters = self.adapters
        units_dict = {"sample": [], "type": [], "platform": [], "machine": [],
                      "processing_unit": [], "barcode": [], "methylation": [],
                      "bam": []}
        if adapters is not None:
            units_dict["adapter"] = []

        file_list = []
        for d in self.directory:
            dir_files_found = 0
            log.info(f"Dir: {d}")
            for f in glob.glob(f"{d}/**/*.bam", recursive=True):
                file_list.append(f)
                dir_files_found += 1
                log.info(f"    found: {f}")
            if dir_files_found == 0:
                log.warning(f"No BAM files found in '{d}' ")
                exit(1)
            else:
                log.info(f"{dir_files_found} BAM files found")

        log.info("Processing found BAM files, extracting run information:")
        for bam_file in file_list:
            log.info(f"\t - {bam_file} for run information")
            rg_dict = extract_bam_information(bam_file, self.default_barcode,
                                              platform)
            sample = rg_dict["sample"]
            units_dict["sample"].append(sample)
            if data_json and 'units' in data_columns:
                extra_columns = list(data_columns['units'].keys())
                if "platform" in extra_columns:
                    value = data_columns['units']['platform']
                    platform = extract_value(value[1][2:], value[0],
                                             data_json['samples'][sample])
                    units_dict["platform"].append(platform)
                else:
                    units_dict["platform"].append(platform)
                if "type" in extra_columns:
                    value = data_columns['units']['type']
                    units_dict["type"] = extract_value(
                        value[1][2:], value[0], data_json['samples'][sample])
                else:
                    units_dict["type"].append(sample_type)
                if "adapters" in extra_columns:
                    value = data_columns['units']['adapter']
                    units_dict["adapter"] = extract_value(
                        value[1][2:], value[0], data_json['samples'][sample])
                else:
                    if adapters is not None:
                        units_dict["adapter"].append(adapters)
            else:
                units_dict["type"].append(sample_type)
                units_dict["platform"].append(platform)
                if adapters is not None:
                    units_dict["adapter"].append(adapters)

            units_dict["machine"].append(rg_dict["machine"])
            units_dict["processing_unit"].append(rg_dict["pu"])
            units_dict["barcode"].append(rg_dict["barcode"])
            units_dict["bam"].append(bam_file)
            units_dict["methylation"].append(rg_dict["methylation"])

            if platform == 'ONT':
                if "basecalling_model" not in units_dict.keys():
                    units_dict["basecalling_model"] = []
                    units_dict["run_id"] = []

                units_dict["basecalling_model"].append(rg_dict["basecalling_model"])
                units_dict["run_id"].append(rg_dict["run_id"])

        units_df = pd.DataFrame(units_dict)

        # Check for duplicated rows which indicate bam files with the same read group info
        # This triggers a warning that these bam files with the same could be merged
        if platform == 'ONT':
            cols_to_check = ["sample", "type", "run_id", "platform", "machine",
                             "processing_unit", "barcode"]
        else:
            cols_to_check = ["sample", "type", "platform", "machine",
                             "processing_unit", "barcode"]
        if units_df[cols_to_check].duplicated().sum() > 1:
            log.warning("The RG information is the same in multiple BAM files."
                        "Consider merging bam files with the same read groups")

        samples_file_name = "samples.tsv"
        if self.post_file_modifier is not None:
            samples_file_name = "samples_{}.tsv".format(
                self.post_file_modifier)
        if os.path.isfile(samples_file_name):
            if not self.force:
                log.warning("File exists {} and force wasn't used".format(
                    samples_file_name))
                exit(1)
            else:
                log.warning("File exists {} overwriting!!!".format(
                    samples_file_name))
        samples_series = units_df["sample"].drop_duplicates()
        samples_df = pd.DataFrame(samples_series)

        if data_columns and "samples" in data_columns:
            extra_cols_header = ["sample"] + \
                list(data_columns["samples"].keys())
        elif self.tc is not None:
            samples_df.insert(1, "tumor_content",
                              [self.tc] * samples_df.shape[0])

        if data_columns and "samples" in data_columns:
            all_rows = []
            for sample in samples_series:
                row_data = [sample]
                for _, value in data_columns['samples'].items():
                    row_data.append(extract_value(value[1][2:], value[0],
                                                  data_json['samples'][sample]))
                all_rows.append(row_data)

            extra_cols_df = pd.DataFrame(all_rows, columns=extra_cols_header)
            samples_df = pd.merge(samples_df, extra_cols_df, on="sample",
                                  validate="one_to_one")

        samples_df.sort_values(by="sample", inplace=True)
        samples_df.to_csv(samples_file_name, index=False, sep='\t')

        units_file_name = "units.tsv"
        if self.post_file_modifier is not None:
            units_file_name = "units_{}.tsv".format(self.post_file_modifier)
        if os.path.isfile(units_file_name):
            if not self.force:
                log.warning("File exists {} and force wasn't used".format(units_file_name))
                exit(1)
            else:
                log.warning("File exists {} overwriting!!!".format(units_file_name))

        if platform == "ONT":
            col_order = ["sample", "type", "platform", "machine", "processing_unit",
                         "run_id", "barcode", "methylation",
                         "basecalling_model", "bam"]
            units_df = units_df[col_order]

        extra_header_columns = []
        if data_columns and 'units' in data_columns:
            extra_header_columns = list(data_columns['units'].keys())
            if 'type' in extra_header_columns:
                extra_header_columns.remove("type")
            if 'platform' in extra_header_columns:
                extra_header_columns.remove("platform")
            if 'adapter' in extra_header_columns:
                extra_header_columns.remove("adapter")

        if len(extra_header_columns) > 0:
            extra_units_cols_header = ["sample"] + extra_header_columns
            extra_cols = []
            extra_columns = list(data_columns['units'].keys())
            for sample in samples_series:
                extra_data = []
                for column in extra_columns:
                    value = data_columns['units'][column]
                    extra_data.append(extract_value(value[1][2:], value[0], data_json['samples'][sample]))
                extra_cols.append(extra_data)

            extra_unit_cols_df = pd.DataFrame(extra_cols, columns=extra_units_cols_header)
            units_df = pd.merge(units_df, extra_unit_cols_df, on="sample", validate="one_to_one")

        units_df.sort_values(by="sample", inplace=True)
        units_df.to_csv(units_file_name, index=False, sep='\t', na_rep='NA')


def extract_value(field, default, data):
    head, *tail = field
    if head not in data:
        return default
    if tail:
        return extract_value(tail, default, data[head])
    else:
        return data[head]


def extract_bam_information(file_path, default_barcode=None, platform="ONT"):

    """
    extract read group information and check for methylation tags
    in the provided BAM file.

    :param file_path: path to BAM file
    :type file_path: string
    :param default_barcode: barcode string used when a barcode can not
    be extracted
    :type default_barcode: string
    :param: platform: specifies which sequencing platform this is
    :type platform: string

    :return: dictionary with keys sample_id, machine_id, processing_unit_id barcode
    Additional basecalling_model and run_id when platform is ONT
    :rtype: dict
    """

    bam = pysam.AlignmentFile(file_path, "rb", check_sq=False)
    header_dict = bam.header.to_dict()
    read_group = header_dict["RG"]
    read_group_num = len(read_group)  # check for for more than one read group in the bam header
    if read_group_num > 1:
        log.warning(f"{read_group_num} read groups found in uBAM {file_path}."
                    "Choosing the first read group for the units.tsv file")

    read_group_dict = read_group[0]

    bam_platform = read_group_dict["PL"]
    if bam_platform != platform:
        log.error(f"The platform specified in the BAM file is not {platform}!")
        exit(1)

    pu = read_group_dict["PU"]

    try:  # PM tag not in some older pacbio BAM files
        machine_id = read_group_dict["PM"]
    except KeyError:
        machine_id = "NA"

    read_group_description = {}
    for desc in read_group_dict['DS'].split(' '):
        desc_fields = desc.split("=")
        read_group_description[desc_fields[0]] = desc_fields[1]

    if platform == "ONT":
        basecall_model = read_group_description["basecall_model"]
        run_id = read_group_description["runid"]

    try:
        barcode = read_group_dict["BC"]
    except KeyError:
        barcode = default_barcode

    # some ONT bam have only LB and it is the sample id.
    try:
        sample_id = read_group_dict["SM"]
    except KeyError:
        sample_id = read_group_dict["LB"]

    # check first read for methylation tags
    first_read = bam.head(1)
    for read in first_read:
        if read.has_tag('MM') and read.has_tag('ML'):
            methylation_tags = 'Yes'
        else:
            methylation_tags = 'No'

    rg_dict = {"sample": sample_id, "machine": machine_id,
               "pu": pu, "barcode": barcode,
               "methylation": methylation_tags}

    if platform == "ONT":
        rg_dict["basecalling_model"] = basecall_model
        rg_dict["run_id"] = run_id

    return rg_dict


def extract_run_information(file_path, default_barcode=None, number_of_reads=200, every_n_reads=1000, warning_threshold=0.9,
                            compare_first_and_last_read=False, ask_for_input=False):
    """
    extract information from provided fastq.gz file and creates a consensus create_barcode

    :param file_path: path to fastq.gz file
    :type file_path: string
    :param default_barcode: barcode string used when a barcode can not be extracted
    :type default_barcode: string
    :param number_of_reads: number of reads that will be used to create consensus barcode
    :type number_of_reads: integer
    :param: warning_threshold: raise a warning for char with lower occurences this value
    :type warning_threshold: float
    :param compare_first_and_last_read: compare first read with last read to detect merged lanes or flowcells

    :return: (machine_id, flowcell_id, lane, consensus_barcode)
    :rtype: tuple
    """

    def extract_barcode(line):
        """
        extract barcode from read name

        :param line: readname, ex @A00687:159:HLCNMDRXY:1:2101:2320:1000:GCC+AGG 1:N:0:CGCTCTAT+TTGCAACG
        :type line: string

        :return: barcode from read name
        :rtype: string
        """
        return re.sub('[12/]+$', '', line.decode("utf-8").split(":")[-1])

    def count_bases(data, barcode, length):
        """
        iterate over barcode and increate data counter

        :param data: data structure representing each position in the barcode,
                     ex [[{'A': 0, 'C': 0, 'G': 0, 'T': 0, 'N': 0, '+': 0} ...]
        :typa data: list of dicts
        :param barcode: barcode added to data structure, ex ACGT+GTCA
        :type barcode: string
        :param length: length of barcode

        :return: updated version of data structure
        :rtype: list of dicts
        """
        for i in range(length):
            if barcode[i] == 'A':
                data[i]['A'] += 1
            elif barcode[i] == 'C':
                data[i]['C'] += 1
            elif barcode[i] == 'G':
                data[i]['G'] += 1
            elif barcode[i] == 'T':
                data[i]['T'] += 1
            elif barcode[i] == 'N':
                data[i]['N'] += 1
            elif barcode[i] == '+':
                data[i]['+'] += 1
        return data

    def extract_run_informatio(line):
        """
        extract information from read name

        :param line: read name, ex @A00687:159:HLCNMDRXY:1:2101:2320:1000:GCC+AGG 1:N:0:CGCTCTAT+TTGCAACG
        :return tuple with (machine_id, flowcell_id, lane)
        :rtype: tuple
        """
        columns = line.decode("utf-8").split(":")
        return (columns[0], columns[2], columns[3])

    def create_barcode(data, length, number_of_reads, warning_threshold=0.9):
        """
        from data a barcode will be generate. Warnings will be raised for consensus chars that have a lower occurences then
        the provided threshold.

        :param data: data structure representing each position in the barcode,
                     ex [[{'A': 0, 'C': 0, 'G': 0, 'T': 0, 'N': 0, '+': 0} ...]
        :typa data: list of dicts
        :param length: length of barcode
        :type length: integer
        :param number_of_reads: number of reads that have been use to create data
        :type number_of_reads: integer
        :param: warning_threshold: raise a warning for char with lower occurences this value
        :type warning_threshold: float

        :return: consensus barcode
        :rtype: string
        """
        barcode = ""
        for i in range(length):
            max_base = max(data[i], key=data[i].get)
            if list(data[i].values()).count(data[i][max_base]) > 1:
                raise Exception("Multiple base with same occurences: {}. UNABLE to handle exiting!".format(data[i]))
            max_base_n = data[i][max_base]
            if max_base_n / number_of_reads < warning_threshold:
                logging.warning('Consensus base {} occurences {:.1%} at position {} in barcode, file {}'.
                                format(max_base, max_base_n / number_of_reads, i, file_path))
            barcode += max_base
        return barcode

    def ask_user_for_input(message, question):
        """
        Function used to ask user for input, retries once if no input is given.
        """
        counter = 2
        user_input = ''
        print(message)
        while counter > 0 and len(user_input) == 0:
            user_input = input(question).rstrip()
            counter -= 1
        if len(user_input) == 0:
            raise Exception("No input entered!!!")
        if '_' in user_input:
            logging.warning("Replacing all occurences of '_' with '-' in {}".format(user_input))
            user_input = user_input.replace("_", "-")
        return user_input

    def skip_read_information(reader_it):
        """
        used to skip read sequence and quality lines
        """
        next(reader_it)
        next(reader_it)
        next(reader_it)

    with gzip.open(file_path, "rb") as reader:
        counter = number_of_reads
        every = every_n_reads  # only look at every n read
        reader_it = iter(reader)
        # Parse first read
        line = next(reader_it).rstrip()
        # Extract machine id, flowcell and lane
        machine_id, flowcell_id, lane = extract_run_informatio(line)
        barcode = extract_barcode(line)
        if not re.match('^[A-Z-+]+$', barcode):
            if default_barcode is not None:
                return (machine_id, flowcell_id, lane, default_barcode)
            else:
                raise Exception("Undable to extract barcode from read name and no default barcode specified")
        skip_read_information(reader_it)
        counter -= 1
        length = len(barcode)
        # data structure used to store counts for each barcode
        data = [{'A': 0, 'C': 0, 'G': 0, 'T': 0, 'N': 0, '+': 0} for i in range(length)]
        data = count_bases(data, barcode, length)
        last_read = line
        for line in reader_it:
            if every == 0:
                every = every_n_reads
                line = line.rstrip()
                last_read = line
                data = count_bases(data, extract_barcode(line), length)
                skip_read_information(reader_it)
                counter -= 1
                if counter == 0:
                    break
                continue
            else:
                line = line.rstrip()
                last_read = line
                skip_read_information(reader_it)
                every -= 1
        if counter > 0:
            logging.warning("Could only select {} reads of {} from fastq file {} for evaluation!".
                            format(number_of_reads - counter, number_of_reads, file_path))
        if compare_first_and_last_read:
            if counter == 0:
                for last_read in reader_it:
                    skip_read_information(reader_it)
            data = count_bases(data, extract_barcode(last_read), length)
            last_machine_id, last_flowcell_id, last_lane = extract_run_informatio(last_read)
            if last_machine_id != machine_id:
                if ask_for_input:
                    last_machine_id = ask_user_for_input("Multiple machines found in fastq file, {} and {}\n".
                                                         format(last_machine_id, machine_id),
                                                         "Enter machine id that should be used:")
                else:
                    raise Exception("Multiple machines found in fastq file, {} and {}".format(last_machine_id, machine_id))
            if last_flowcell_id != flowcell_id:
                if ask_for_input:
                    last_flowcell_id = ask_user_for_input("Multiple flowcells found in fastq file, {} and {}".
                                                          format(last_flowcell_id, flowcell_id),
                                                          "Enter flowcell id that should be used:")
                else:
                    raise Exception("Multiple flowcells found in fastq file, {} and {}".format(last_flowcell_id, flowcell_id))
            if last_lane != lane:
                logging.warning("First read and last read have different lane numbers {} vs {}, lane will be set to 0!".
                                format(last_lane, lane))
            return (last_machine_id, last_flowcell_id, "0", create_barcode(data, length, number_of_reads - counter, warning_threshold))
        return (machine_id, flowcell_id, lane, create_barcode(data, length, number_of_reads - counter, warning_threshold))



#!/usr/bin/env python
"""
    Main hydra_genetics module file.

"""
from packaging.version import Version
from . import _version
__version__ = _version.get_versions()['version']


def min_version(version):
    """Require minimum hydra-genetics version, raise workflow error if not met."""
    if Version(__version__) < Version(version):
        from .exceptions import HydraGeneticsVersionError
        raise HydraGeneticsVersionError(
            "Expecting Hydra-Genetics version {} or higher (you are currently using {}).".format(
                version, __version__
            )
        )


def max_version(version):
    """Set maximum hydra-genetics version, raise workflow error if not met."""
    if Version(__version__) > Version(version):
        from .exceptions import HydraGeneticsVersionError
        raise HydraGeneticsVersionError(
            "Expecting Hydra-Genetics version {} or less (you are currently using {}).".format(
                version, __version__
            )
        )



# -*- coding: UTF-8 -*-

from __future__ import print_function

__author__ = "Patrik Smeds"
__copyright__ = "Copyright 2022, Patrik Smeds"
__email__ = "patrik.smeds@scilifelab.uu.se"
__license__ = "GPL-3"


from pathlib import Path
import sys
import versioneer

if sys.version_info < (3, 8):
    print("At least Python 3.8 is required for hydra-genetics.\n", file=sys.stderr)
    exit(1)

try:
    from setuptools import setup, find_packages
except ImportError:
    print("Please install setuptools before installing hydra-genetics.", file=sys.stderr)
    exit(1)

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="hydra-genetics",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Helper tools for use with hydra-genetics pipelines.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=[
        "hydra-genetics",
        "snakemake",
        "bioinformatics",
        "workflow",
        "pipeline",
        "clinical",
        "biology",
        "sequencing",
        "NGS",
        "next generation sequencing",
    ],
    author="Patrik Smeds",
    author_email="patrik.smeds@scilifelab.uu.se",
    url="https://github.com/hydra-genetics/tools",
    license='GPL-3',
    entry_points={"console_scripts": ["hydra-genetics=hydra_genetics.__main__:run"]},
    install_requires=[
        'pandas>=1.3.1',
        'click>=8,<9',
        'jinja2==3.0.1',
        'rich==10.9.0',
        'snakemake',
        'pysam',
        'gitpython',
        'pyaml',
        'packaging'
    ],
    packages=find_packages(exclude=("docs")),
    include_package_data=True,
    zip_safe=False,
)



import git
import hashlib
import logging
import os
import re
import subprocess
import yaml

from datetime import datetime
from snakemake.common import is_local_file


def _create_container_name_version_string(image_information):
    if image_information.endswith(".sif"):
        container_name_and_version = re.search(r"__([A-Za-z0-9-]+)_([A-Za-z0-9.-]+)\.sif$", image_information)
        if container_name_and_version:
            return "_".join(container_name_and_version.groups())
        else:
            container_name = re.search(r"([A-Za-z0-9-]+)\.sif$", image_information)
            if container_name:
                return "_".join([container_name.groups()[0], "NoVersion"])
            else:
                raise Exception(f"Unable to extract container name from {image_information}")
    else:
        container_name_and_version = re.search("/([A-Za-z0-9-_.]+):[ ]*([A-Za-z0-9-_.]+)$", image_information)
        if container_name_and_version:
            return "_".join(container_name_and_version.groups())
        else:
            container_name = re.search("/([A-Za-z0-9-]+)$", image_information)
            if container_name:
                return "_".join([container_name.groups()[0], "NoVersion"])
            else:
                raise Exception(f"Unable to extract container name from {image_information}")


def add_version_files_to_multiqc(config, file_list):
    if "multiqc" in config.keys():
        for report in config["multiqc"]["reports"]:
            config["multiqc"]["reports"][report]["qc_files"] += file_list
    else:
        logger = logging.getLogger(__name__)
        logger.warning(f"Could not find key multiqc in configuration.")


def _touch(fname):
    if os.path.exists(fname):
        os.utime(fname, None)
    else:
        open(fname, "a").close()


def get_container_prefix(workflow):
    """
    Function used to fetch singularity cache location
    """

    if hasattr(workflow, "use_singularity"):
        # Fetch singularity prefix for snakemake version with version less
        # then 8
        if hasattr(workflow, "singularity_prefix"):
            return workflow.singularity_prefix
        else:
            return ".snakemake/singularity"
    elif hasattr(workflow, "deployment_settings"):
        # Fetch singularity prefix for snakemake version with equal to 8.0 or newer
        if workflow.deployment_settings.apptainer_prefix is None:
            return ".snakemake/singularity"
        else:
            return workflow.deployment_settings.apptainer_prefix
    return None


def get_software_version_from_labels(image_path):
    """
    Function used to create a list of software version used in a singularity image,
    usually created from a docker image. The software name and version will be extracted
    from labels that have been set in the container.  The expected format is
    ([A-Za-z0-9-_.]+): ([0-9.]+)$, where the first section is the software name
    and second part is the version. In the dockerfile the labels could look like this
    LABEL gatk4=4.11.0

    Parameters
    ----------
    image_path: str
       path to the image that will be inspected

    Returns:
    --------
    list of tuples where each tuple consist of software name and software version.

    """
    cmd = ["singularity", "inspect", image_path]
    software_version_list = []
    for row in subprocess.check_output(cmd).decode().split("\n"):

        if not row.startswith("org.") and not row.startswith("maintainer") and not row.startswith("license"):
            software_version = re.match("^([A-Za-z0-9-_.]+): ([a-z0-9.]+)$", row)
            if software_version:
                software_version_list.append(software_version.groups())
    return software_version_list


def get_deffile(image_path):
    """
    Function used to extract from which docker image the singularity image was created.

    Parameters
    ----------
    image_path: str
       path to the image that will be inspected

    Returns
    -------
    docker image name "org/name:version or None if information couldn't be found.

    """
    cmd = ["singularity", "inspect", image_path]
    for row in subprocess.check_output(cmd).decode().split("\n"):
        if row.startswith("org.label-schema.usage.singularity.deffile.from"):
            return row.split(": ")[1]
    return None


def get_image_path(container, singularity_prefix_path):
    """
    Function used to return a path to singularity a image. If it's a local file, .i.e
    a path to a file on the file system the path will be returned without modifications. If it's
    a singularity image that has been downloaded by snakemake the function will create
    the snakemake generated image name and path, which is the singularity prefix as storage
    location followed by a hashname (generated from the provided docker image name and location).

    Parameters
    ----------
    container: str
        either a path to a local file or image name and type, ex docker://hydragenetics/bcbio-vc:0.2.6
    :singularity_prefix_path: str
        storage path of singularity/apptainer images

    Returns
    -------
    str: path to locally store image
    """
    if is_local_file(container):
        return container

    md5hash = hashlib.md5()
    md5hash.update(container.encode())
    return os.path.join(singularity_prefix_path, f"{md5hash.hexdigest()}.simg")


def use_container(workflow):
    """
    Function used to check if containers are used,
    """
    if hasattr(workflow, "use_singularity"):
        # For snakemake with version less then 8
        return True
    elif hasattr(workflow, "deployment_settings") and workflow.deployment_settings.deployment_method:
        # For snakemake with version 8 or newer
        from snakemake.settings import DeploymentMethod

        if DeploymentMethod.APPTAINER in workflow.deployment_settings.deployment_method:
            return True
        else:
            return False
    else:
        return False


def add_software_version_to_config(config, workflow, fail_missing_versions=True):
    """
    Function that will look through the config dict to locate
    container definitions. When container information is found the
    softwares version will be extracted and added to the config.

    Parameters:
    config: dict
        snakemake dict object
    workflow: object
        snakemake workflow object
    fail_missing_versions: boolean
        if set to true an error will be raised if a container is specified
        and versions can't be extracted.

    Returns:
    a config where software_verions have been added
    """
    container_cache = get_container_prefix(workflow)

    def _add_software_version(config, version_dict):
        logger = logging.getLogger(__name__)
        version_found = []
        software_version_key = "software_versions"
        for key, value in config.items():
            if isinstance(value, dict):
                config[key], version_dict = _add_software_version(value, version_dict)
            elif key in ["container", "default_container"]:
                if key == "default_container":
                    software_version_key = f"default_container_software_versions"
                image_path = get_image_path(value, container_cache)
                if os.path.isfile(image_path):
                    version_found += get_software_version_from_labels(image_path)
                    name_and_version = _create_container_name_version_string(value)

                    if not version_found:
                        if fail_missing_versions:
                            raise Exception(f"could not extract software versions from {image_path}, {value}")
                        else:
                            logger.warning(f"could not extract software versions from {image_path}, {value}")
                            name = name_and_version.split("_")[0]
                            version = "_".join(name_and_version.split("_")[1:])
                            version_found = [[name, version], ("NOTE", "version extract from image name and not labels")]
                else:
                    if fail_missing_versions:
                        raise Exception(f"could not locate local file {image_path} for {value}")
                    else:
                        logger.warning(f"could not locate local file {image_path} for {value}")
        if version_found:
            config[software_version_key] = {s: v for s, v in version_found}
            if "version" in config[software_version_key]:
                del config[software_version_key]["version"]
            version_dict[name_and_version] = config[software_version_key]
        return config, version_dict

    return _add_software_version(config, {})


def export_software_version_as_file(
    software_dict, directory="versions/software", file_name="softwares_mqc_versions.yaml", date_string=None
):
    """
    Print software version to file (should be same as filename in touch_software_version_file). Requires a dict with key software_info which
    should contain a dict with software names and versions.

    Parameters:
    -----------
    software_dict: dict
       dict with key software_info, ex {software_info: {'common_1.3': {'samtools': '1.3', 'picard': '1.6'}}}
    directory: str
       path where files will be written
    file_name: str
       file name to use for software logging
    date_string: str
       a string that will be added to the folder name to make it unique
    """
    if date_string is None:
        date_string = datetime.now().strftime("%Y%m%d")
    if directory is not None and len(directory) > 0:
        date_string = f"_{date_string}"
    directory = f"{directory}{date_string}"
    if len(directory) > 0 and not os.path.isdir(directory):
        os.makedirs(directory)
    output_file = os.path.join(directory, f"{file_name}")
    with open(output_file, "w") as writer:
        writer.write(yaml.dump(software_dict))
        for name in software_dict:
            notification = software_dict[name].pop("NOTE", None)
            if notification is not None:
                writer.write(f"# {notification}")
    return output_file


def touch_software_version_file(config, directory="versions/software", file_name="softwares_mqc_versions.yaml", date_string=None):
    """
    To be able to pass a proper file list to multiqc the version files need to exist before the dag graph
    is built. And before the dag graph is built we can not guarantee that all images exist on the file system,
    and therefor we can not use export_software_version_as_file to directly create the actual version files.
    This function will create all files (empty) that export_software_version_as_file will create.
    NOTE: you need to have the same parameter as with export_software_version_as_file

    Parameters:
    -----------
    config: dict
       dict with config file # not used?
    directory: str
       path where files will be written
    file_name: str
       file name to use for software version logging
    date_string: str
       a string that will be added to the folder name to make it unique
    """

    if date_string is None:
        date_string = datetime.now().strftime("%Y%m%d")
    if directory is not None and len(directory) > 0:
        date_string = f"_{date_string}"
        directory = f"{directory}{date_string}"
    if len(directory) > 0 and not os.path.isdir(directory):
        os.makedirs(directory)
    output_file = os.path.join(directory, f"{file_name}")
    _touch(output_file)

    if not file_name.endswith("mqc_versions.yaml"):
        logger = logging.getLogger(__name__)
        logger.warning(f"{file_name} is not going to be recognized by MultiQC. Should end with mqc_versions.yaml")
    return output_file


def get_pipeline_version(workflow, pipeline_name="pipeline"):
    """
    Will return the pipelines tag name and commit hex.

    Parameters:
    -----------
    workflow: object
        workflow object from snakemake that contain a basedir attribute

    Return
    ------
    dict with pipeline version and commit, ex {'pipeline_name': {'version': 'v1', 'pipeline_commit': 'achgk2...kacaa'}}
    """
    logger = logging.getLogger(__name__)

    def _find_root_repo(path):
        if path is None or os.path.dirname(str(path)) == str(path):
            return None
        elif os.path.isdir(str(os.path.join(str(path), ".git"))):
            return path
        else:
            return _find_root_repo(os.path.dirname(path))

    repo_path = _find_root_repo(getattr(workflow, "basedir"))

    if not repo_path:
        logger.warning("Pipeline directory is not a git repo")
        return {pipeline_name: {"version": None, "commit_id": None}}

    # Initialize a Git repo object
    repo = git.Repo(repo_path)

    # Get the currently checked out commit
    head_commit = repo.head.commit

    # Iterate through all tags and find the one pointing to the HEAD commit
    pipeline_version = None
    for tag in repo.tags:
        if tag.commit == head_commit:
            pipeline_version = tag.name
            break
    if pipeline_version is None:
        try:
            pipeline_version = repo.active_branch
        except TypeError as e:
            logger.warning("Unable to get version (from tags) or an active_branch")

    return {pipeline_name: {"version": str(pipeline_version), "commit_id": str(head_commit)}}


def export_pipeline_version_as_file(
    pipeline_version_dict, directory="versions/software", file_name_ending="mqc_versions.yaml", date_string=None
):
    """
    Print pipeline version to a file. Requires a dict with key pipeline_info which
    should contain a dict with software names and versions.

    Parameters:
    -----------
    pipeline_version_dict: dict
       dict with key pipeline_version_dict, ex {pipeline_name: {'version': '1.3', 'commit_id': 'ac8ac8t73'}}
    directory: str
       path where files will be written
    file_name_ending: str
       a string that will be combined with the software name version key
    date_string: str
       a string that will be added to the folder name to make it unique
    """
    if date_string is None:
        date_string = datetime.now().strftime("%Y%m%d")
    if directory is not None and len(directory) > 0:
        date_string = f"_{date_string}"
    directory = f"{directory}{date_string}"
    if len(directory) > 0 and not os.path.isdir(directory):
        os.makedirs(directory)
    output_file_list = []
    for pipeline_name in pipeline_version_dict:
        output_file = os.path.join(
            directory, f"{pipeline_name}__{pipeline_version_dict[pipeline_name]['version']}_{file_name_ending}"
        )
        output_file_list.append(output_file)
        with open(output_file, "w") as writer:
            writer.write(yaml.dump({pipeline_name: pipeline_version_dict[pipeline_name]["version"]}))
    return output_file_list


def touch_pipeline_version_file_name(
    pipeline_version_dict, directory="versions/software", file_name_ending="mqc_versions.yaml", date_string=None
):
    """
    Function used to create empty version file for pipeline version. This
    function is a bit redundant and export_pipeline_version_as_file could be used,
    since this function isn't dependent on singularity images which
    export_software_version_as_file is. But the function exist to have a similar workflow
    for both pipeline and software. NOTE: you need to have same parameters into
    touch_pipeline_version_file_name as to export_pipeline_version_as_file.

    Parameters:
    -----------
    pipeline_version_dict: dict
       dict with key pipeline_version_dict, ex {pipeline_name: {'version': '1.3', 'commit_id': 'ac8ac8t73'}}
    directory: str
       path where files will be written
    file_name_ending: str
       a string that will be combined with the software name version key
    date_string: str
       a string that will be added to the folder name to make it unique
    """
    if date_string is None:
        date_string = datetime.now().strftime("%Y%m%d")
    if directory is not None and len(directory) > 0:
        date_string = f"_{date_string}"
    directory = f"{directory}{date_string}"
    if len(directory) > 0 and not os.path.isdir(directory):
        os.makedirs(directory)

    output_file_list = []
    for pipeline_name in pipeline_version_dict:
        output_file = os.path.join(
            directory, f"{pipeline_name}__{pipeline_version_dict[pipeline_name]['version']}_{file_name_ending}"
        )
        output_file_list.append(output_file)
        _touch(output_file)
    return output_file_list



import os
import git
import tempfile
import unittest
from unittest.mock import Mock

from hydra_genetics.utils import software_versions


class TestPipelineVersion(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_plain_directory_version(self):
        # create a mock pipeline, in this case an empty directory
        pipeline_dir = tempfile.mkdtemp()

        # create a mock workflow
        workflow = Mock(basedir=pipeline_dir)

        version = software_versions.get_pipeline_version(workflow, "test_pipeline")
        self.assertEqual(version, {"test_pipeline": {"version": None, "commit_id": None}})

    def test_git_directory_version(self):
        # create a mock repo with an empty README
        repo_dir = tempfile.mkdtemp()
        repo = git.Repo.init(repo_dir, initial_branch="develop")
        repo_file = os.path.join(repo_dir, "README.md")
        open(repo_file, "wb").close()
        repo.index.add([repo_file])
        commit = repo.index.commit("initial commit")

        # create a mock workflow
        workflow = Mock(basedir=repo_dir)

        # no tag, so version will be the branch name
        version = software_versions.get_pipeline_version(workflow, "test_pipeline")
        self.assertEqual(version, {"test_pipeline": {"version": "develop", "commit_id": commit.hexsha}})

        # create a tag
        repo.create_tag("v1.2.3")
        version = software_versions.get_pipeline_version(workflow, "test_pipeline")
        self.assertEqual(version, {"test_pipeline": {"version": "v1.2.3", "commit_id": commit.hexsha}})



# vim: syntax=python tabstop=4 expandtab
# coding: utf-8

__author__ = "{{ author }}"
__copyright__ = "Copyright {{ year }}, {{ author }}"
__email__ = "{{ email }}"
__license__ = "GPL-3"


def dummy():
    return 1
